"""
Modifiers provide convenient mechanisms for modifying and combining
:class:`revscoring.Feature` and constant values into new
:class:`revscoring.Feature`.

.. autofunction:: revscoring.features.modifiers.log

----

.. autofunction:: revscoring.features.modifiers.min
.. autofunction:: revscoring.features.modifiers.max

----

.. autofunction:: revscoring.features.modifiers.add
.. autofunction:: revscoring.features.modifiers.sub
.. autofunction:: revscoring.features.modifiers.mul
.. autofunction:: revscoring.features.modifiers.div

----

.. autofunction:: revscoring.features.modifiers.eq
.. autofunction:: revscoring.features.modifiers.ne
.. autofunction:: revscoring.features.modifiers.gt
.. autofunction:: revscoring.features.modifiers.lt
.. autofunction:: revscoring.features.modifiers.ge
.. autofunction:: revscoring.features.modifiers.le

"""
from .feature import (add, and_, div, eq, ge, gt, le, log, lt, max, min, mul,
                      ne, not_, or_, sub)

__all__ = [add, div, eq, ge, gt, le, log, lt, max, min, mul, ne, sub, and_,
           or_, not_]



"""
This module contains a collection of models that implement a simple function:
:func:`~revscoring.ScorerModel.score`.  Currently, all models are
a subclass of :class:`revscoring.scorer_models.MLScorerModel`
which means that they also implement
:meth:`~revscoring.scorer_models.MLScorerModel.train` and
:meth:`~revscoring.scorer_models.MLScorerModel.test` methods.

Support Vector Classifiers
++++++++++++++++++++++++++
.. automodule:: revscoring.scorer_models.svc

Naive Bayes Classifiers
+++++++++++++++++++++++
.. automodule:: revscoring.scorer_models.nb

Random Forest
+++++++++++++
.. automodule:: revscoring.scorer_models.rf

Abstract classes
++++++++++++++++
.. automodule:: revscoring.scorer_models.scorer_model


"""
from .svc import SVC, SVCModel, LinearSVC, LinearSVCModel, RBFSVC, RBFSVCModel
from .nb import (NB, NBModel, GaussianNB, GaussianNBModel, MultinomialNB,
                 MultinomialNBModel, BernoulliNB, BernoulliNBModel)
from .scorer_model import ScorerModel, MLScorerModel
from .sklearn_classifier import ScikitLearnClassifier
from .rf import RF, RFModel

__all__ = [
    SVC, SVCModel, LinearSVC, LinearSVCModel, RBFSVC, RBFSVCModel,
    NB, NBModel, GaussianNB, GaussianNBModel, MultinomialNB,
    MultinomialNBModel, BernoulliNB, BernoulliNBModel,
    ScorerModel, MLScorerModel, ScikitLearnClassifier,
    RF, RFModel
]



from nose.tools import eq_

from ..dependent import Dependent


def test_dependent():

    foobar1 = Dependent("foobar", lambda: "foobar1")
    foobar2 = Dependent("foobar", lambda: "foobar2")

    eq_(foobar1, foobar2)

    eq_(hash(foobar1), hash(foobar2))

    assert foobar1 in {foobar2}



import pickle

from nose.tools import eq_

from .. import turkish
from ...datasources import revision
from ...dependencies import solve
from .util import compare_extraction

BAD = [
    "adamın dib", "adamın dip",
    "ahlaksız",
    "ahmak",
    "allahsız",
    "am", "amcık",
    "amk", "amq",
    "amın oğlu",
    "amına", "amına koy", "amına koyayım", "amına koyyim",
    "amını",
    "ananı",
    "ananın am", "ananın dölü",
    "ananızın", "ananızın am",
    "anasını",
    "anasının am",
    "antisemitic",
    "aptal",
    "asdf",
    "ağzına sıçayım",
    "beyinsiz",
    "bi bok", "bok", "boktan", "bokça",
    "dedeler",
    "dinci", "dinsiz",
    "dönek",
    "dıcks",
    "embesil",
    "eshek",
    "gerizekalı",
    "gerzek",
    "godoş",
    "gotten",
    "göt", "göt deliği", "göt oğlanı", "götlek", "götoğlanı", "götveren",
    "götü", "götün",
    "haysiyetsiz",
    "heval", "hewal",
    "huur",
    "i.b.n.e", "ibne", "ibnedir", "ibneli k", "ibnelik",
    "inci",
    "israil köpektir",
    "it oğlu it",
    "kaltak",
    "kaşar",
    "kevaşe",
    "kıç",
    "liboş",
    "mal",
    "meme",
    "nesi kaşar",
    "nobrain",
    "o. çocuğ",
    "orospu", "orospu cocugu", "orospu çoc", "orospu çocuğu",
    "orospu çocuğudur", "orospudur", "orospunun", "orospunun evladı",
    "orospuçocuğu",
    "oğlan", "oğlancı",
    "pezeven", "pezeveng", "pezevengin evladı", "pezevenk",
    "pisliktir",
    "piç",
    "puşt", "puşttur",
    "pıgs",
    "reyiz",
    "sahip",
    "serkan",
    "salak",
    "sik", "sikem", "siken", "siker", "sikerim", "sikey", "sikici", "sikik",
    "sikil", "sikiş", "sikişme", "sikm", "sikseydin", "sikseyidin",
    "sikt", "siktim", "siktir", "siktir lan",
    "sokarım", "sokayım",
    "swicht şamandra",
    "tipini s.k", "tipinizi s.keyim",
    "veled", "weled",
    "woltağym",
    "woğtim",
    "wulftim",
    "yarrak", "yarrağ",
    "yavş", "yavşak",
    "yavşaktır",
    "zippo dünyanın en boktan çakmağıdır",
    "zortlamasi",
    "zıkkımım",
    "zıonısm",
    "çük",
    "ıbnelık",
    "ın tröst we trust",
    "şerefsiz",
]

INFORMAL = [
    "achtırma",
    "achıyorlardı",
    "arkadashlar",
    "arkadashım",
    "basharamıyor",
    "basharan",
    "basharıla",
    "bashka",
    "bashlangıc",
    "bashlıyor",
    "bashıma",
    "bashının",
    "beshinci",
    "beshtane",
    "chalıshmıshlar", "chalıshıldıgında", "chalıshılınırsa", "chalıshıyorlar",
    "chalıshıyorsun", "chalıshıyorum",
    "chamashırlar",
    "charpmadan",
    "charptı", "charpınca",
    "chevirsin", "chevırdım",
    "chimdirecem",
    "choluk",
    "chorusdan", "choruslar",
    "chıka", "chıkacak", "chıkmadı", "chıksa", "chıktı", "chıkıp", "chıkısh",
    "chıplak",
    "degishen", "degishiklik", "degishiyor",
    "dönüshü",
    "eastblacksea",
    "eshekoglu",
    "eshkıyanın",
    "gardash", "gardashlık", "gardaslık",
    "gecherken",
    "gechirdi", "gechirmeyeyeyim", "gechiyor", "gechti", "gechtim",
    "genish",
    "gerchegi",
    "ichimde", "ichimden", "ichin", "ichine", "ichini",
    "ichlerinde",
    "ishler",
    "kalmısh",
    "kardeshlerini", "kardeshlerinizde",
    "karshı",
    "keshke",
    "kishiler", "kishilerin", "kishinin", "kishiye", "kishiyle",
    "konushmalrını", "konushmuyor", "konushuldugunda", "konushur",
    "koshan", "koshtu", "koshuyor",
    "nıshanlı",
    "saatchide",
    "sachlı",
    "sanatchılarını",
    "shansin", "shansın",
    "shashırtma", "shashırtmaya",
    "sheytanlar",
    "takabeg",
    "theır",
    "uchakdan",
    "uzanmıshsın",
    "yaklashdıkdan",
    "yakıshıklı",
    "yaratılmısh",
    "yashamak", "yashatıyorsun", "yashta", "yashıyorsun",
    "yerleshtiriyorum",
    "yozgatfm",
    "üch",
    "ıchın",
    "ıshıksız"
]

OTHER = [
    """
    Türkiye'deki en üst seviye futbol ligi. Bir sezonda 18 takımın mücadele
    ettiği ligde, her takımın diğerleriyle ikişer maç yaptığı çift devreli lig
    usulü uygulanmaktadır. En üst sırada yer alan takım şampiyon olurken son üç
    sıradaki takım 1. Lig'e düşmekte, 1. Lig'deki üç takım ise ertesi sezon
    mücadele etmek üzere Süper Lig'e yükselmektedir.
    """
]


def test_badwords():
    compare_extraction(turkish.revision.badwords_list, BAD, OTHER)


def test_informals():
    compare_extraction(turkish.revision.informals_list, INFORMAL, OTHER)


def test_revision():
    # Words
    cache = {revision.text: "Bir sezonda m18, takımın mücadele."}
    eq_(solve(turkish.revision.words_list, cache=cache),
        ["Bir", "sezonda", "m18", "takımın", "mücadele"])


def test_pickling():

    eq_(turkish, pickle.loads(pickle.dumps(turkish)))



from datetime import datetime

from pytz import utc

from . import modifiers
from ..datasources import revision
from .feature import Feature
# TODO: Many of these are enwiki specific -- they shouldn't be.
from .util import (CATEGORY_RE, CITE_RE, IMAGE_RE, INFOBOX_RE, MARKUP_RE,
                   NUMERIC_RE, SECTION_COMMENT_RE, SYMBOLIC_RE)


# ############################### Time ########################################

def process_day_of_week(revision_metadata):

    dt = datetime.fromtimestamp(revision_metadata.timestamp.unix(), tz=utc)
    return dt.weekday()

day_of_week = Feature("revision.day_of_week", process_day_of_week,
                      returns=int, depends_on=[revision.metadata])
"""
Represents day of week when the edit was made in UTC.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [revision.day_of_week]))
        [6]
"""


def process_hour_of_day(revision_metadata):

    dt = datetime.fromtimestamp(revision_metadata.timestamp.unix(), tz=utc)
    return dt.hour

hour_of_day = Feature("revision.hour_of_day", process_hour_of_day,
                      returns=int, depends_on=[revision.metadata])
"""
Represents hour of day when the edit was made in UTC.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [revision.hour_of_day]))
        [21]
"""


# ############################### Comment #####################################
def process_has_custom_comment(revision_metadata):

    if revision_metadata.comment is not None:
        trimmed_comment = SECTION_COMMENT_RE.sub("", revision_metadata.comment)
        trimmed_comment = trimmed_comment.strip()

        return len(trimmed_comment) > 1
    else:
        return False

has_custom_comment = Feature("revision.has_custom_comment",
                             process_has_custom_comment,
                             returns=bool, depends_on=[revision.metadata])
"""
Represents whether the edit has custom edit summary.

:Returns:
    bool

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [revision.has_custom_comment]))
        [False]
"""


def process_has_section_comment(revision_metadata):

    if revision_metadata.comment is not None:
        return SECTION_COMMENT_RE.match(revision_metadata.comment) is not None
    else:
        return False

has_section_comment = \
    Feature("revision.has_section_comment", process_has_section_comment,
            returns=bool, depends_on=[revision.metadata])
"""
Represents whether the edit has section edit summary.

Section edit summaries are like ``"/* <section title> */"`` which mediawiki
automatically creates them when a user edit a section.

:Returns:
    bool

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [revision.has_section_comment]))
        [True]
"""


# ################################ Bytes ######################################
def process_bytes(revision_metadata):
    return revision_metadata.bytes or 0

bytes = Feature("revision.bytes", process_bytes,
                returns=int, depends_on=[revision.metadata])
"""
Represents size of the revision's content in bytes. It uses UTF-8 encoding.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [revision.bytes]))
        [24028]
"""


# ############################### Characters ##################################
def process_chars(revision_text):
    return len(revision_text)

chars = Feature("revision.chars", process_chars,
                returns=int, depends_on=[revision.text])
"""
Represents number of characters in the revision.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [revision.chars]))
        [24016]
"""


def process_markup_chars(revision_text):
    return sum(len(m.group(0)) for m in MARKUP_RE.finditer(revision_text))

markup_chars = Feature("revision.markup_chars", process_markup_chars,
                       returns=int, depends_on=[revision.text])
"""
Represents number of markup characters in the revision.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [revision.markup_chars]))
        [704]
"""

proportion_of_markup_chars = markup_chars / modifiers.max(chars, 1)
"""
Represents ratio of markup characters compared to all characters in revision.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130,
        ...      [revision.proportion_of_markup_chars]))
        [0.029313790806129246]
"""


def process_numeric_chars(revision_text):
    return sum(len(m.group(0)) for m in NUMERIC_RE.finditer(revision_text))

numeric_chars = Feature("revision.numeric_chars", process_numeric_chars,
                        returns=int, depends_on=[revision.text])
"""
Represents number of numeric characters in the revision.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [revision.numeric_chars]))
        [203]
"""

proportion_of_numeric_chars = numeric_chars / modifiers.max(chars, 1)
"""
Represents ratio of numeric characters compared to all characters in revision.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130,
        ...      [revision.proportion_of_numeric_chars]))
        [0.008452698201199201]
"""


def process_symbolic_chars(revision_text):
    return sum(len(m.group(0)) for m in SYMBOLIC_RE.finditer(revision_text))

symbolic_chars = Feature("revision.symbolic_chars",
                         process_symbolic_chars,
                         returns=int, depends_on=[revision.text])
"""
Represents number of symbolic characters in the revision.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [revision.symbolic_chars]))
        [2559]
"""

proportion_of_symbolic_chars = symbolic_chars / modifiers.max(chars, 1)
"""
Represents ratio of symbolic characters compared to all characters in revision.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130,
        ...      [revision.proportion_of_symbolic_chars]))
        [0.10655396402398401]
"""


def process_uppercase_chars(revision_text):
    return sum(c.lower() != c for c in revision_text)

uppercase_chars = Feature("revision.uppercase_chars",
                          process_uppercase_chars,
                          returns=int, depends_on=[revision.text])
"""
Represents number of uppercase characters in the revision.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [revision.uppercase_chars]))
        [742]
"""

proportion_of_uppercase_chars = uppercase_chars / modifiers.max(chars, 1)
"""
Represents ratio of uppercase characters compared to all characters in
revision.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130,
        ...      [revision.proportion_of_uppercase_chars]))
        [0.030896069287141906]
"""


# ############################## Parse tree ###################################
def process_level_1_headings(headings):
    return sum(h.level == 1 for h in headings)

level_1_headings = \
    Feature("revision.level_1_headings", process_level_1_headings,
            returns=int, depends_on=[revision.headings])
"""
Represents number of first level headings in the revision.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [revision.level_1_headings]))
        [0]
"""


def process_level_2_headings(headings):
    return sum(h.level == 2 for h in headings)

level_2_headings = \
    Feature("revision.level_2_headings", process_level_2_headings,
            returns=int, depends_on=[revision.headings])
"""
Represents number of second level headings in the revision.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [revision.level_2_headings]))
        [6]
"""


def process_level_3_headings(headings):
    return sum(h.level == 3 for h in headings)

level_3_headings = \
    Feature("revision.level_3_headings", process_level_3_headings,
            returns=int, depends_on=[revision.headings])
"""
Represents number of third level headings in the revision.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [revision.level_3_headings]))
        [13]
"""


def process_level_4_headings(headings):
    return sum(h.level == 4 for h in headings)

level_4_headings = \
    Feature("revision.level_4_headings", process_level_4_headings,
            returns=int, depends_on=[revision.headings])
"""
Represents number of forth level headings in the revision.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [revision.level_4_headings]))
        [4]
"""


def process_level_5_headings(headings):
    return sum(h.level == 5 for h in headings)

level_5_headings = \
    Feature("revision.level_5_headings", process_level_5_headings,
            returns=int, depends_on=[revision.headings])
"""
Represents number of fifth level headings in the revision.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [revision.level_5_headings]))
        [0]
"""


def process_level_6_headings(headings):
    return sum(h.level == 5 for h in headings)

level_6_headings = \
    Feature("revision.level_6_headings", process_level_6_headings,
            returns=int, depends_on=[revision.headings])
"""
Represents number of sixth level headings in the revision.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [revision.level_6_headings]))
        [0]
"""


def process_content_chars(content):
    return len(content)

content_chars = Feature("revision.content_chars", process_content_chars,
                        returns=int, depends_on=[revision.content])
"""
Represents number of content characters in the revision.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [revision.content_chars]))
        [19363]
"""


def process_internal_links(revision_internal_links):
    return len(revision_internal_links)

internal_links = Feature("revision.internal_links", process_internal_links,
                         returns=int, depends_on=[revision.internal_links])
"""
Represents number of internal links in the revision.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130, [revision.internal_links]))
        [146]
"""


def process_image_links(revision_internal_links):
    return sum(1 for l in revision_internal_links
               if IMAGE_RE.match(str(l.title)))

image_links = Feature("revision.image_links", process_image_links,
                      returns=int, depends_on=[revision.internal_links])
"""
Represents number of image links in the revision.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(661258898, [revision.image_links]))
        [13]
"""


def process_category_links(revision_internal_links):
    return sum(1 for l in revision_internal_links
               if CATEGORY_RE.match(str(l.title)))

category_links = Feature("revision.category_links", process_category_links,
                         returns=int, depends_on=[revision.internal_links])
"""
Represents number of category links in the revision.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(661258898, [revision.category_links]))
        [50]
"""


def ref_tags_process(revision_tags):
    return sum(1 for tag in revision_tags if tag.tag == "ref")

ref_tags = Feature("revision.ref_tags", ref_tags_process, returns=int,
                   depends_on=[revision.tags])
"""
Represents number of reference tags in the revision.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(661258898, [revision.ref_tags]))
        [228]
"""


def process_templates(revision_templates):
    return len(revision_templates)

templates = Feature("revision.templates", process_templates,
                    returns=int, depends_on=[revision.templates])
"""
Represents number of templates in the revision.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(661258898, [revision.templates]))
        [287]
"""


def process_cite_templates(revision_templates):
    return sum(1 for t in revision_templates if CITE_RE.search(str(t.name)))

cite_templates = Feature("revision.cite_templates", process_cite_templates,
                         returns=int, depends_on=[revision.templates])
"""
Represents number of citation templates (e.g. "Cite web") in the revision.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(661258898, [revision.cite_templates]))
        [175]
"""


proportion_of_templated_references = \
    cite_templates / modifiers.max(ref_tags, 1)
"""
Represents ratio of number of citation templates compared to number of
reference tags in the revision.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(655097130,
        ...      [revision.proportion_of_templated_references]))
        [0.7272727272727273]
"""


def process_infobox_templates(revision_templates):
    return sum(1 for t in revision_templates if INFOBOX_RE.search(str(t.name)))

infobox_templates = Feature("revision.infobox_templates",
                            process_infobox_templates,
                            returns=int, depends_on=[revision.templates])
"""
Represents number of infobox templates (e.g. "Infobox scientist") in the
revision.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import revision
        >>> list(extractor.extract(661258898, [revision.infobox_templates]))
        [1]
"""

all = [day_of_week, hour_of_day,
       has_custom_comment, has_section_comment,
       bytes,
       chars, markup_chars, numeric_chars, symbolic_chars, uppercase_chars,
       proportion_of_markup_chars, proportion_of_numeric_chars,
       proportion_of_symbolic_chars, proportion_of_uppercase_chars,
       level_1_headings, level_2_headings, level_3_headings, level_4_headings,
       level_5_headings, level_6_headings, internal_links,
       image_links, category_links, ref_tags, templates, cite_templates,
       infobox_templates]



"""
.. autoclass:: revscoring.Datasource
    :member-order:
    :inherited-members:
"""
from ..dependencies import Dependent


class Datasource(Dependent):
    """
    Represents a data source for generating features.  Unlike features,
    datasources do not necessarily generate simple scalar values.

    :Parameters:
        name : str
            The name of the feature
        process : `func`
            A function that will generate a data value
        depends_on : `list`(`hashable`)
            An ordered list of dependencies that correspond
            to the `*args` of `process`
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __hash__(self):
        return hash(('datasource', self.name))



import sys

from .space_delimited import SpaceDelimited

try:
    import enchant
    dictionary = enchant.Dict("vi")
except enchant.errors.DictNotFoundError:
    raise ImportError("No enchant-compatible dictionary found for 'vi'.  " +
                      "Consider installing 'hunspell-vi'.")

# https://vi.wiktionary.org/wiki/Th%C3%A0nh_vi%C3%AAn:Laurent_Bouvier/
# Free_Vietnamese_Dictionary_Project_Vietnamese-Vietnamese#Allwiki_.28closed.29
stopwords = set([
    "ai", "bằng", "bị", "bộ", "cho", "chưa", "chỉ", "cuối", "cuộc",
    "các", "cách", "cái", "có", "cùng", "cũng", "cạnh", "cả", "cục",
    "của", "dùng", "dưới", "dừng", "giữa", "gì", "hay", "hoặc",
    "khi", "khác", "không", "luôn", "là", "làm", "lại", "mà", "mọi",
    "mỗi", "một", "nhiều", "như", "nhưng", "nào", "này", "nữa",
    "phải", "qua", "quanh", "quá", "ra", "rất", "sau", "sẽ", "sự",
    "theo", "thành", "thêm", "thì", "thứ", "trong", "trên", "trước",
    "trừ", "tuy", "tìm", "từng", "và", "vài", "vào", "vì", "vẫn",
    "về", "với", "xuống", "đang", "đã", "được", "đấy", "đầu", "đủ"
])
badwords = [
    # Vietnamese
    r"[ck]ặ[tc]", r"[ck]u", r"cứt", r"(dz?|gi)âm", r"đái", r"đéo", r"đ[ụù]",
    r"đĩ", r"đ[íị]t", r"ỉa", r"l[ôồ]n", r"trứng"
]
informals = [
    # Vietnamese
    r"bợn", r"bro",
    r"chẳng", r"ch[ớứ]", r"cú",
    r"đụ", r"đừng", r"fải",
    r"khỉ",
    r"mày", r"nghịch", r"ngu", r"ngụy", r"nguỵ",
    r"ok", r"ơi",
    r"quái",
    r"thằng", r"thôi", r"tui", r"ừ", r"vời", r"wái?",
    r"zì"
]

sys.modules[__name__] = SpaceDelimited(
    __name__,
    doc="""
vietnamese
==========

revision
--------
.. autoattribute:: revision.words
.. autoattribute:: revision.content_words
.. autoattribute:: revision.badwords
.. autoattribute:: revision.misspellings
.. autoattribute:: revision.informals

parent_revision
---------------
.. autoattribute:: parent_revision.words
.. autoattribute:: parent_revision.content_words
.. autoattribute:: parent_revision.badwords
.. autoattribute:: parent_revision.misspellings
.. autoattribute:: parent_revision.informals

diff
----
.. autoattribute:: diff.words_added
.. autoattribute:: diff.words_removed
.. autoattribute:: diff.badwords_added
.. autoattribute:: diff.badwords_removed
.. autoattribute:: diff.informals_added
.. autoattribute:: diff.informals_removed
.. autoattribute:: diff.misspellings_added
.. autoattribute:: diff.misspellings_removed
    """,
    badwords=badwords,
    dictionary=dictionary,
    informals=informals,
    stopwords=stopwords
)
"""
vietnamese
----------
"""



import sys

from .space_delimited import SpaceDelimited

try:
    from nltk.stem.snowball import SnowballStemmer
    stemmer = SnowballStemmer("french")
except ValueError:
    raise ImportError("Could not load stemmer for {0}. ".format(__name__))

try:
    from nltk.corpus import stopwords as nltk_stopwords
    stopwords = set(nltk_stopwords.words('french') + ["a"])
except LookupError:
    raise ImportError("Could not load stopwords for {0}. ".format(__name__) +
                      "You may need to install the nltk 'stopwords' " +
                      "corpora.  See http://www.nltk.org/data.html")

try:
    import enchant
    dictionary = enchant.Dict("fr")
except enchant.errors.DictNotFoundError:
    raise ImportError("No enchant-compatible dictionary found for 'fr'.  " +
                      "Consider installing 'myspell-fr'.")

badwords = [
    r"con",
    r"fesse", r"foutre",
    r"merde+", r"merdique",
    r"prostituee?", r"putain", r"putes",
    r"salop", r"stupide",
]

sys.modules[__name__] = SpaceDelimited(
    __name__,
    doc="""
french
======

revision
--------
.. autoattribute:: revision.words
.. autoattribute:: revision.content_words
.. autoattribute:: revision.badwords
.. autoattribute:: revision.misspellings
.. autoattribute:: revision.infonoise

parent_revision
---------------
.. autoattribute:: parent_revision.words
.. autoattribute:: parent_revision.content_words
.. autoattribute:: parent_revision.badwords
.. autoattribute:: parent_revision.misspellings
.. autoattribute:: parent_revision.infonoise

diff
----
.. autoattribute:: diff.words_added
.. autoattribute:: diff.words_removed
.. autoattribute:: diff.badwords_added
.. autoattribute:: diff.badwords_removed
.. autoattribute:: diff.misspellings_added
.. autoattribute:: diff.misspellings_removed
    """,
    badwords=badwords,
    dictionary=dictionary,
    stemmer=stemmer,
    stopwords=stopwords
)



import pickle
import traceback

from ..errors import (CaughtDependencyError, DependencyError, DependencyLoop,
                      MissingResource, RevisionNotFound)


def test_exceptions_picklability():
    de = DependencyError("FooBar")
    pickle.loads(pickle.dumps(de))

    try:
        assert False, "woops"
    except Exception as e:
        tb = traceback.extract_stack()
        cde = CaughtDependencyError("FooBar", e, tb)
        pickled_cde = pickle.dumps(cde)
        pickle.loads(pickled_cde)

    dl = DependencyLoop("FooBar")
    pickle.loads(pickle.dumps(dl))

    mr = MissingResource("FooBar")
    pickle.loads(pickle.dumps(mr))

    rnf = RevisionNotFound()
    pickle.loads(pickle.dumps(rnf))



"""
Tunes a set of models against a training set to identify the best
model/configuration.

Usage:
    tune <params-config> <features> [--observations=<path>]
                                    [--scoring=<type>]
                                    [--test-prop=<prop>]
                                    [--folds=<num>]
                                    [--report=<path>]
                                    [--label-type=<type>]
                                    [--processes=<num>]
                                    [--verbose]
                                    [--debug]

Options:
    <params-config>        The path to a YAML configuration file containing the
                           models and parameter values to search when tuning
    <features>             The classpath to a feature_list to use when
                           interpreting the feature values of the observations
    --observations=<path>  The path to a file containing observations to train
                           and test against. [default: <stdin>]
    --scoring=<type>       The type of scoring strategy to optimize for when
                           choosing parameter sets [default: roc_auc]
    --folds=<num>          The number of cross-validation folds to try
                           [default: 5]
    --report=<path>        Path to a file to write the tuning report to
                           [default: <stdout>]
    --label-type=<type>    A type describing the value to expect as a label
                           [default: str]
    --processes=<num>      The number of parallel processes to start for
                           model building [default: <cpu-count>]
    --verbose              Print progress information to stderr
    --debug                Print debug information to stderr

"""
import datetime
import json
import logging
import multiprocessing
import sys
import time
import traceback
from collections import defaultdict

import docopt
import yamlconf
from sklearn import cross_validation, grid_search
from tabulate import tabulate

from . import util
from .. import __version__

logger = logging.getLogger(__name__)


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)

    logging.basicConfig(
        level=logging.INFO if not args['--debug'] else logging.DEBUG,
        format='%(asctime)s %(levelname)s:%(name)s -- %(message)s'
    )

    params_config = yamlconf.load(open(args['<params-config>']))

    features_path = args['<features>']
    features = yamlconf.import_module(features_path)

    label_decoder = util.DECODERS[args['--label-type']]
    if args['--observations'] == "<stdin>":
        observations_f = sys.stdin
    else:
        observations_f = open(args['--observations'])

    observations = util.read_observations(observations_f, features,
                                          label_decoder)

    scoring = args['--scoring']
    folds = int(args['--folds'])

    if args['--report'] == "<stdout>":
        report = sys.stdout
    else:
        report = open(args['--report'], "w")

    if args['--processes'] == "<cpu-count>":
        processes = multiprocessing.cpu_count()
    else:
        processes = int(args['--processes'])

    verbose = args['--verbose']

    run(params_config, features_path, observations, scoring, folds,
        report, processes, verbose)


def run(params_config, features_path, observations, scoring, folds,
        report, processes, verbose):

    observations = list(observations)

    # Prepare the worker pool
    logger.debug("Starting up multiprocessing pool (processes={0})"
                 .format(processes))
    pool = multiprocessing.Pool(processes=processes)

    # Start writing the model tuning report
    possible_labels = set(label for _, label in observations)
    report.write("# Model tuning report\n")
    report.write("- Revscoring version: {0}\n".format(__version__))
    report.write("- Features: {0}\n".format(features_path))
    report.write("- Date: {0}\n".format(datetime.datetime.now().isoformat()))
    report.write("- Observations: {0}\n".format(len(observations)))
    report.write("- Labels: {0}\n".format(json.dumps(list(possible_labels))))
    report.write("- Scoring: {0}\n".format(scoring))
    report.write("- Folds: {0}\n".format(folds))
    report.write("\n")

    # For each estimator and paramset, submit the job.
    cv_result_sets = defaultdict(lambda: [])
    for name, estimator, param_grid in _estimator_param_grid(params_config):
        logger.debug("Submitting jobs for {0}:".format(name))
        for params in param_grid:
            logger.debug("\tsubmitting {0}..."
                         .format(format_params(params)))
            result = pool.apply_async(_cross_validate,
                                      [observations, estimator, params],
                                      {'scoring': scoring, 'folds': folds})
            cv_result_sets[name].append((params, result))

    # Barrier synchronization
    logger.info("Running gridsearch for {0} model/params pairs ..."
                .format(sum(len(p_r) for p_r in cv_result_sets)))
    grid_scores = []
    for name, param_results in cv_result_sets.items():
        for params, result in param_results:
            scores = result.get()  # This is a line that blocks
            grid_scores.append((name, params, scores.mean(), scores.std()))

    # Write the rest of the report!  First, print the top 10 combinations
    report.write("# Top scoring configurations\n")
    grid_scores.sort(key=lambda gs: gs[2], reverse=True)
    table = tabulate(
        ((name, round(mean_score, 3), round(scores.std(), 3),
          format_params(params))
         for name, params, mean_score, scores in
         grid_scores[:10]),
        headers=["model", "mean(scores)", "std(scores)", "params"]
    )
    report.write(table + "\n")
    report.write("\n")

    # Now print out scores for each model.
    report.write("# Models\n")
    for name, param_results in cv_result_sets.items():
        report.write("## {0}\n".format(name))

        param_scores = ((p, r.get()) for p, r in param_results)
        param_stats = [(p, s.mean(), s.std()) for p, s in param_scores]
        param_stats.sort(key=lambda v: v[1], reverse=True)

        table = tabulate(
            ((round(mean_score, 3), round(scores.std(), 3),
              format_params(params))
             for params, mean_score, scores in
             param_stats),
            headers=["mean(scores)", "std(scores)", "params"]
        )
        report.write(table + "\n")
        report.write("\n")

    report.close()


def format_params(doc):
    return ", ".join("{0}={1}".format(k, json.dumps(v))
                     for k, v in doc.items())


def _estimator_param_grid(params_config):
    for name, config in params_config.items():
        try:
            EstimatorClass = yamlconf.import_module(config['class'])
            estimator = EstimatorClass()
        except Exception:
            logger.warn("Could not load estimator {0}"
                        .format(config['class']))
            logger.warn("Exception:\n" + traceback.format_exc())
            continue

        if not hasattr(estimator, "fit"):
            logger.warn("Estimator {0} does not have a fit() method."
                        .format(config['class']))
            continue

        param_grid = grid_search.ParameterGrid(config['params'])

        yield name, estimator, param_grid


def _cross_validate(observations, estimator, params, scoring="roc_auc",
                    folds=5, verbose=False):

    start = time.time()
    feature_values, labels = (list(vect) for vect in zip(*observations))
    estimator.set_params(**params)

    try:
        scores = cross_validation.cross_val_score(
            estimator, feature_values, labels, scoring=scoring, cv=folds)

        duration = time.time() - start
        logging.debug("Cross-validated {0} with {1} in {2} hours: {3} ({4})"
                      .format(estimator.__class__.__name__,
                              format_params(params),
                              round(duration / (60 * 60), 3),
                              round(scores.mean(), 3),
                              round(scores.std(), 3)))
        return scores

    except Exception:
        logger.warn("Could not load estimator {0}"
                    .format(estimator.__class__.__name__))
        logger.warn("Exception:\n" + traceback.format_exc())
        return [0]*folds



import pickle

from nose.tools import eq_

from .. import spanish
from ...datasources import revision
from ...dependencies import solve
from .util import compare_extraction

BAD = [
    "aweonao",
    "awevo",
    "babosa",
    "boludo",
    "buseta",
    "cabron", "cabrón", "cabrones", "cabrones",
    "caca",
    "caga", "cagada", "cagado",
    "cago", "cagon", "cagones",
    "caquita",
    "carajo",
    "chichornia",
    "chiguero",
    "chimar",
    "chinga", "chingada", "chingadazo", "chingaderita", "chingaderitas",
    "chingado", "chingados", "chingoncisimo", "chingonería",
    "chingones", "chingonsicimo", "chingonsisimo", "chingorrón", "chinguen",
    "chinoncicimo",
    "chosto",
    "chupame", "chupamea", "chupamela", "chupan", "chupar", "chupenmela",
    "cochetumadre",
    "cojio", "cojones",
    "coño",
    "culero", "culeros",
    "culiaos", "culito",
    "cupan",
    "desmadrado", "desmadrar", "desmadrarse", "desmadre",
    "dlaversh",
    "emputado",
    "enc", "encabronadas", "encabronado", "encabronar", "encabronarse",
    "encronada",
    "enputado",
    "follaban", "follar",
    "fregon", "fregón",
    "gamberro",
    "gilipollas",
    "golfa",
    "guevo", "guevon", "guevones",
    "guey", "güey",
    "güila",
    "hifueputa",
    "hijodeputa", "hijodeputas", "hijodputa", "hijodputas", "hijoduta",
    "huevon", "huevones", "huevospateados",
    "idiota", "idiotas",
    "imbecil", "imbécil", "imbeciles", "imbesil",
    "jetear",
    "jidiota",
    "jilipollas", "jilipuertas",
    "joder", "jodido",
    "jotas", "joterias",
    "joto", "jotos",
    "komekaka",
    "lamadre",
    "madreado", "madrear",
    "malparida", "malparidos",
    "mamada", "mamadas", "mamaguevo", "mamon", "mamon", "mamón", "mamones",
    "marica", "maricas", "maricon", "maricones", "marika",
    "marrano",
    "meachupan",
    "meapelan",
    "mecagoenlaleche",
    "mecastralamadre",
    "melachupan", "melapelan",
    "merda", "merga",
    "mevalemergas", "mevalevergas",
    "mierda", "mierdas",
    "mimsn",
    "mojon",
    "monda",
    "nemames", "nememes", "nomames",
    "ogt", "ogts",
    "ojete",
    "pamearlo",
    "pajaro",
    "pattaya",
    "pedos",
    "pelan",
    "pelotudo", "pelotudos",
    "pendejada", "pendejadas",
    "pendejo", "pendejos",
    "pene", "penes",
    "perras",
    "petardas", "petardo",
    "picha", "pichula",
    "pija", "pijas",
    "pinche", "pinches",
    "pirobos",
    "pitito", "pito", "pitos", "pitote",
    "polla", "pollas", "pollas",
    "poronga",
    "puta", "putamadre", "putas",
    "putisima", "putisimo", "putiza",
    "puto", "puton", "putos",
    "sorete",
    "tajodido",
    "tetas",
    "tontos", "tonta",
    "trolo",
    "uta", "utama", "utamadre",
    "verga", "vergas",
    "vergasos", "vergazos", "verguda", "vergudo",
    "versh",
    "violo",
    "watdafuq",
    "webon", "webonada", "webones",
    "yjodeputa",
    "zorra", "zorrear",
]

INFORMAL = [
    "aki",
    "amaona",
    "amigui",
    "amo", "amooo", "amoooo",
    "asi",
    "atte",
    "bla", "blablabla", "blah", "blahblahblah",
    "bobos",
    "bubis",
    "chafa",
    "chale",
    "chí",
    "chido",
    "chilear", "chiliar",
    "esq",
    "estupida", "estúpida", "estupidas", "estupides", "estupido", "estupido",
    "estupidos", "estúpidos",
    "grasias",
    "grax",
    "holaaaaa", "holi", "holis", "holis", "hoooola", "hooooli",
    "ijos",
    "inserta",
    "jaja", "jajaja", "jajajaja", "jeje", "jejejeje",
    "ke",
    "kiero",
    "komo",
    "loco", "loko", "lokos",
    "lol", "lolololololololololol", "looooooool",
    "madrazo",
    "mensos",
    "metia",
    "migis", "miguis",
    "muxo",
    "nocheto",
    "nooooo",
    "nop",
    "oooli", "oooola",
    "osea",
    "pipi",
    "plis",
    "popo",
    "porfavor", "por favor",
    "porque", "por que", "porke", "porq", "porqe",
    "porquería", "porquerías",
    "profe",
    "pupu",
    "qiero",
    "shí",
    "sierto",
    "siiii",
    "soi",
    "tambn", "tmbn",
    "tkm",
    "tqm",
    "vallanse",
    "wena", "weno",
    "wey",
    "XD", "xdd", "xddddd"
]

OTHER = [
    """
    Su cuerpo es largo y estilizado, de un color gris parduzco, menos en su
    parte inferior, que es blanquecina. Existen dos subespecies diferenciadas:
    el rorcual del norte, que tiene su hábitat en el Atlántico Norte, y el
    rorcual antártico, de mayor tamaño, que vive habitualmente en aguas del
    océano Antártico. Puede verse en los principales océanos del planeta,
    desde las aguas polares a las tropicales.
    """,
]


def test_badwords():
    compare_extraction(spanish.revision.badwords_list, BAD, OTHER)


def test_informals():
    compare_extraction(spanish.revision.informals_list, INFORMAL, OTHER)


def test_revision():
    # Words
    cache = {revision.text: "Su cuerpo es largo y estilizado, está áreas."}
    eq_(solve(spanish.revision.words_list, cache=cache),
        ["Su", "cuerpo", "es", "largo", "y", "estilizado", "está", "áreas"])

    # Misspellings
    cache = {revision.text: 'Su cuerpo es largo y worngly. <td>'}
    eq_(solve(spanish.revision.misspellings_list, cache=cache), ["worngly"])

    # Infonoise
    cache = {revision.text: "Su cuerpo es largo!"}
    eq_(solve(spanish.revision.infonoise, cache=cache), 9/15)


def test_pickling():

    eq_(spanish, pickle.loads(pickle.dumps(spanish)))



"""
.. autoclass:: revscoring.Dependent
    :members:
"""
import logging

logger = logging.getLogger(__name__)


def not_implemented(*args, **kwargs):
    raise NotImplementedError("Not implemented.")


class Dependent:
    """
    Constructs a dependency-handling processor function.

    :Parameters:
        name : str
            A name to identify this dependency
        process : func
            A function to run when solving this dependency
        depends_on : `iterable`
            A collection of :class:`revscoring.Dependent` whose values are
            required by `process`
    """
    def __init__(self, name, process=None, depends_on=None,
                 dependencies=None):
        self.name = name
        self.process = process or not_implemented
        self.dependencies = dependencies or depends_on or []
        self.calls = 0

    def __call__(self, *args, **kwargs):
        logger.debug("Executing {0} ({1} calls so far)."
                     .format(self, self.calls))
        self.calls += 1
        return self.process(*args, **kwargs)

    def __hash__(self):
        return hash(('dependent', self.name))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<" + self.name + ">"



from nose.tools import eq_

from ..context import Context
from ..dependent import Dependent


def test_context():
    # No context
    context = Context()
    foo = Dependent("foo", lambda: "foo")
    bar = Dependent("bar", lambda: "bar")
    foobar = Dependent("foobar", lambda foo, bar: foo + bar,
                       depends_on=[foo, bar])
    eq_(context.solve(foobar), "foobar")
    eq_(list(context.solve([foo, bar, foobar])), ["foo", "bar", "foobar"])

    # Cache context
    context = Context(cache={bar: "baz"})
    eq_(context.solve(foobar), "foobaz")

    # Context context
    mybar = Dependent("bar", lambda: "baz")

    context = Context(context={mybar})
    eq_(context.solve(foobar), "foobaz")

    context = Context(context={mybar: mybar})
    eq_(context.solve(foobar), "foobaz")

    context = Context(context={bar: mybar})
    eq_(context.solve(foobar), "foobaz")

    context = Context(context={bar: lambda: "baz"})
    eq_(context.solve(foobar), "foobaz")
